#
# https = //w3c.github.io/webdriver/webdriver-spec.html
# WebDriver Protocol Implemenation
#

from base64 import b64decode

from retrying import retry

from .asserters import is_displayed
from .command import Command
from .locator import Locator
from .remote_invoker import RemoteInvoker
from .util import add_element_extension_method, value_to_key_strokes, fluent
from .webdriverresult import WebDriverResult
from .webdriverexception import WebDriverException
from .webelement import WebElement


class WebDriver(object):
    """The WebDriver Object to implement most part of WebDriver protocol.

    Attributes:
        session_id(str): A UDID used to uniquely identify each session.
        capabilities(dict): The capabilities of the driver/browserreturned by
            remote server.
        desired_capabilities(dict): The desired capabilities requested by the
            local end.
        remote_invoker(RemoteInvoker): The remote invoker responsible for send
            request.
    """

    def __init__(self, desired_capabilities, url='http://127.0.0.1:3456/wd/hub'):
        """Initialize the WebDriver

        Args:
            desired_capabilities(dict): The desired capabilities requested by
                the local end.
            url(str): The url of remote server, default: localhost:3456/wd/hub.
        """
        self.session_id = None
        self.capabilities = None
        self.desired_capabilities = desired_capabilities
        self.remote_invoker = RemoteInvoker(url)

    def __repr__(self):
        return '<{0.__name__} (session="{1}")>'.format(
            type(self), self.session_id)

    def _execute(self, command, data=None, unpack=True):
        """ Private method to execute command.

        Args:
            command(Command): The defined command.
            data(dict): The uri variable and body.
            uppack(bool): If unpack value from result.

        Returns:
            The unwrapped value field in the json response.
        """
        if not data:
            data = {}
        if self.session_id is not None:
            data.setdefault('session_id', self.session_id)
        data = self._wrap_el(data)
        res = self.remote_invoker.execute(command, data)
        ret = WebDriverResult.from_object(res)
        ret.raise_for_status()
        ret.value = self._unwrap_el(ret.value)
        if not unpack:
            return ret
        return ret.value

    def _unwrap_el(self, value):
        """Convert {'Element': 1234} to WebElement Object

        Args:
            value(str|list|dict): The value field in the json response.

        Returns:
            The unwrapped value.
        """
        if isinstance(value, dict) and 'ELEMENT' in value:
            element_id = value.get('ELEMENT')
            return WebElement(element_id, self)
        elif isinstance(value, list) and not isinstance(value, str):
            return [self._unwrap_el(item) for item in value]
        else:
            return value

    def _wrap_el(self, value):
        """Convert WebElement Object to {'Element': 1234}

        Args:
            value(str|list|dict): The local value.

        Returns:
            The wrapped value.
        """
        if isinstance(value, dict):
            return {k: self._wrap_el(v) for k, v in value.items()}
        elif isinstance(value, WebElement):
            return {'ELEMENT': value.element_id}
        elif isinstance(value, list) and not isinstance(value, str):
            return [self._wrap_el(item) for item in value]
        else:
            return value

    @property
    def sessions(self):
        """Gets all the sessions of the webdriver server.

        Support:
            Android iOS Web(WebView)

        Returns:
            Return the URL of the current page.
        """
        return self._execute(Command.GET_ALL_SESSIONS)

    @fluent
    def attach(self, session_id):
        """Attach to given Session.

        Support:
            Android iOS Web(WebView)

        Args:
            session_id(str): The given session ID

        Returns:
            WebDriver Object.
        """
        self.session_id = session_id

    @fluent
    def init(self):
        """Create Session by desiredCapabilities

        Support:
            Android iOS Web(WebView)

        Returns:
            WebDriver Object.
        """
        resp = self._execute(Command.NEW_SESSION, {
            'desiredCapabilities': self.desired_capabilities
        }, False)
        resp.raise_for_status()
        self.session_id = str(resp.session_id)
        self.capabilities = resp.value

    @fluent
    def quit(self):
        """Quit the driver.

        Support:
            Android iOS Web(WebView)

        Returns:
            WebDriver Object.
        """
        self._execute(Command.QUIT)

    @fluent
    def get(self, url):
        """Loads a web page in the current browser session.

        Support:
            Web(WebView)

        Args:
            url(str): The URL to navigate to.

        Returns:
            WebDriver Object.

        Raises:
            WebDriverException.
        """
        self._execute(Command.GET, {'url': url})

    @property
    def current_url(self):
        """Gets the URL of the current page.

        Support:
            Web(WebView)

        Returns:
            Return the URL of the current page.
        """
        return self._execute(Command.GET_CURRENT_URL)

    @fluent
    def back(self):
        """Back.

        Support:
            Web(WebView)

        Returns:
            WebDriver Object.
        """
        self._execute(Command.GO_BACK)

    @fluent
    def forward(self):
        """Forward.

        Support:
            Web(WebView)

        Returns:
            WebDriver Object.
        """
        self._execute(Command.GO_FORWARD)

    @fluent
    def refresh(self):
        """Refresh.

        Support:
            Web(WebView)

        Returns:
            WebDriver Object.
        """
        self._execute(Command.REFRESH)

    @property
    def title(self):
        """Get the current page title.

        Support:
            Web(WebView)
        """
        return self._execute(Command.GET_TITLE)

    @property
    def current_window_handle(self):
        """Returns the handle of the current window.

        Support:
            Web(WebView)
        """
        return self._execute(Command.GET_CURRENT_WINDOW_HANDLE)

    @fluent
    def close(self):
        """Closes the current window.

        Support:
            Web(WebView)

        Returns:
            WebDriver Object.
        """
        self._execute(Command.CLOSE)

    @fluent
    def switch_to_window(self, window_name):
        """Switch to the given window.

        Support:
            Web(WebView)

        Args:
            window_name(str): The window to change focus to.

        Returns:
            WebDriver Object.
        """
        data = {
            'name': window_name
        }
        self._execute(Command.SWITCH_TO_WINDOW, data)

    @property
    def window_handles(self):
        """Returns the handles of all windows within the current session.

        Support:
            Web(WebView)
        """
        return self._execute(Command.GET_WINDOW_HANDLES)

    @fluent
    def maximize_window(self):
        """Maximizes the current window.

        Support:
            Web(WebView)

        Returns:
            WebDriver Object.
        """
        self._execute(Command.MAXIMIZE_WINDOW,
            {"window_handle": "current"})

    @fluent
    def set_window_size(self, width, height, window_handle='current'):
        """Sets the width and height of the current window.

        Support:
            Web(WebView)

        Args:
            width(int): the width in pixels.
            height(int): the height in pixels.
            window_handle(str): Identifier of window_handle,
                default to 'current'.

        Returns:
            WebDriver Object.
        """
        self._execute(Command.SET_WINDOW_SIZE, {
            'width': int(width),
            'height': int(height),
            'window_handle': window_handle})

    def get_window_size(self, window_handle='current'):
        """Gets the width and height of the current window.

        Support:
            Web(WebView)

        Args:
            window_handle(str): Identifier of window_handle,
                default to 'current'.

        Returns:
            A dict contains width and height
        """
        return self._execute(Command.GET_WINDOW_SIZE,
            {'window_handle': window_handle})

    @fluent
    def set_window_position(self, x, y, window_handle='current'):
        """Sets the x,y position of the current window.

        Support:
            Web(WebView)

        Args:
            x(int): the x-coordinate in pixels.
            y(int): the y-coordinate in pixels.
            window_handle(str): Identifier of window_handle,
                default to 'current'.

        Returns:
            WebDriver Object.
        """
        self._execute(Command.SET_WINDOW_POSITION, {
            'x': int(x),
            'y': int(y),
            'window_handle': window_handle})

    def get_window_position(self, window_handle='current'):
        """Gets the x,y position of the current window.

        Support:
            Web(WebView)

        Args:
            window_handle(str): Identifier of window_handle,
                default to 'current'.

        Usage:
            driver.get_window_position()
        """
        return self._execute(Command.GET_WINDOW_POSITION, {
            'window_handle': window_handle})

    @property
    def context(self):
        """returns the current context (Native or WebView).

        Support:
            Android iOS
        """
        return self._execute(Command.CURRENT_CONTEXT_HANDLE)

    @property
    def contexts(self):
        """returns a list of available contexts

        Support:
            Android iOS
        """
        return self._execute(Command.CONTEXT_HANDLES)

    @context.setter
    def context(self, new_context):
        """sets the current context

        Support:
            Android iOS
        """
        self._execute(Command.SWITCH_TO_CONTEXT, {"name": new_context})

    @fluent
    def move_to(self, element, x=0, y=0):
        """Move the mouse by an offset of the specificed element.

        Support:
            Android

        Args:
            element(WebElement): WebElement Object.
            x(float): X offset to move to, relative to the
                      top-left corner of the element.
            y(float): Y offset to move to, relative to the
                      top-left corner of the element.

        Returns:
            WebDriver object.
        """
        self._execute(Command.MOVE_TO, {
            'element': element.element_id,
            'x': x,
            'y': y
        })

    @fluent
    def flick(self, element, x, y, speed):
        """Flick on the touch screen using finger motion events.
           This flickcommand starts at a particulat screen location.

        Support:
            iOS

        Args:
            element(WebElement): WebElement Object where the flick starts.
            x(float}: The x offset in pixels to flick by.
            y(float): The y offset in pixels to flick by.
            speed(float) The speed in pixels per seconds.

        Returns:
            WebDriver object.
        """
        self._execute(Command.FLICK, {
            'element': element.element_id,
            'x': x,
            'y': y,
            'speed': speed
        })

    @fluent
    def tap(self, element):
        """Single tap on the touch enabled device.

        Support:
            Android iOS

        Args:
            element(WebElement): WebElement Object to single tap on.

        Returns:
            WebDriver object.
        """
        self._execute(Command.SINGLE_TAP, {
            'element': element.element_id,
        })

    @fluent
    def swipe(self, startX, startY, endX, endY, duration = 1000):
        """Swipe on the touch screen using finger motion events.

        Support:
            Android iOS

        Args:
            startX(Float): The x-coordinate of start point.
            startY(Float): The y-coordinate of start point.
            endX(Float): The x-coordinate of end point.
            endY(Float): The y-coordinate of end point.
            duration(Float): The duration in ms.

        Returns:
            WebDriver object.
        """
        self._execute(Command.SWIPE_ELEMENT, {
            'element_id': 'temp',
            'startX': startX,
            'startY': startY,
            'endX': endX,
            'endY': endY,
            'duration': duration
        })

    @fluent
    def send_keys(self, value):
        """Send a sequence of key strokes.

        Support:
            Android iOS Web(WebView)

        Args:
            value(str|int|list): value can be a string,
              int or a list contains defined Keys.
        """
        self._execute(Command.SEND_KEYS_TO_ACTIVE_ELEMENT, {
            'value': value_to_key_strokes(value)
        })

    @fluent
    def switch_to_frame(self, frame_reference=None):
        """Switches focus to the specified frame, by index, name, or webelement.

        Support:
            Web(WebView)

        Args:
            frame_reference(None|int|WebElement):
                The identifier of the frame to switch to.
                None means to set to the default context.
                An integer representing the index.
                A webelement means that is an (i)frame to switch to.
                Otherwise throw an error.

        Returns:
            WebDriver Object.
        """
        if frame_reference is not None and type(frame_reference) not in [int, WebElement]:
            raise TypeError('Type of frame_reference must be None or int or WebElement')
        self._execute(Command.SWITCH_TO_FRAME,
            {'id': frame_reference})

    @fluent
    def switch_to_parent_frame(self):
        """Switches focus to the parent context.

        Support:
            Web(WebView)
        """
        self._execute(Command.SWITCH_TO_PARENT_FRAME)


    def get_active_element(self):
        """Returns the active element in current context.

        Support:
            Web(WebView)
        """
        return self._execute(Command.GET_ACTIVE_ELEMENT)

    @property
    def source(self):
        """Gets the source of the current page.

        Support:
            Android iOS Web(WebView)

        Returns:
            Return the source of the current page.
        """
        return self._execute(Command.GET_PAGE_SOURCE)


    def execute_script(self, script, *args):
        """Execute JavaScript Synchronously in current context.

        Support:
            Web(WebView)

        Args:
            script: The JavaScript to execute.
            *args: Arguments for your JavaScript.

        Returns:
            Returns the return value of the function.
        """
        return self._execute(Command.EXECUTE_SCRIPT, {
            'script': script,
            'args': list(args)})

    @fluent
    def execute_async_script(self, script, *args):
        """Execute JavaScript Asynchronously in current context.

        Support:
            Web(WebView)

        Args:
            script: The JavaScript to execute.
            *args: Arguments for your JavaScript.

        Returns:
            Returns the return value of the function.
        """
        return self._execute(Command.EXECUTE_ASYNC_SCRIPT, {
            'script': script,
            'args': list(args)})

    @property
    def cookies(self):
        """Returns  all cookies associated with the address of the current context.

        Support:
            Web(WebView)
        """
        return self._execute(Command.GET_ALL_COOKIES)

    def get_cookie(self, name):
        """Get a single cookie by name.

        Support:
            Web(WebView)

        Returns:
            Returns the cookie if found, None if not.
        """
        for cookie in self.cookies:
            if cookie['name'] == name:
                return cookie
        return None

    @fluent
    def delete_cookie(self, name):
        """Delete a single cookie by name.

        Support:
            Web(WebView)

        Args:
            name: The cookie name.
        """
        self._execute(Command.DELETE_COOKIE, {'name': name})

    @fluent
    def delete_all_cookies(self):
        """Delete all cookies at once.

        Support:
            Web(WebView)
        """
        self._execute(Command.DELETE_ALL_COOKIES)

    @fluent
    def add_cookie(self, cookie_dict):
        """Set a cookie.

        Support:
            Web(WebView)

        Args:
            cookie_dict: A dictionary contain keys: "name", "value",
                ["path"], ["domain"], ["secure"], ["httpOnly"], ["expiry"].

        Returns:
            WebElement Object.
        """
        if not isinstance(cookie_dict, dict):
            raise TypeError('Type of the cookie must be a dict.')
        if not cookie_dict.get(
            'name', None
        ) or not cookie_dict.get(
            'value', None):
            raise KeyError('Missing required keys, \'name\' and \'value\' must be provided.')
        self._execute(Command.ADD_COOKIE, {'cookie': cookie_dict})

    @fluent
    def set_implicitly_wait(self, time_to_wait):
        """Implicit wait timeout that specifies a time to
           wait for the implicit element location strategy
           hen locating elements using Find Element and Find Elements.

        Support:
            Android iOS Web(WebView)

        Args:
            time_to_wait(int): Amount of time to wait (in seconds)

        Returns:
            WebElement Object.
        """
        self._execute(Command.IMPLICIT_WAIT, {
            'ms': float(time_to_wait) * 1000})

    @fluent
    def set_script_timeout(self, time_to_wait):
        """Script timeout that specifies a time to wait for scripts to run.

        Support:
            Web(WebView)

        Args:
            time_to_wait(int): Amount of time to wait (in seconds)

        Returns:
            WebElement Object.
        """
        self._execute(Command.SET_SCRIPT_TIMEOUT, {
            'ms': float(time_to_wait) * 1000})

    @fluent
    def set_page_load_timeout(self, time_to_wait):
        """Page load timeout that specifies a time to
           wait for the page loading to complete.

        Support:
            Web(WebView)

        Args:
            time_to_wait(int): Amount of time to wait

        Returns:
            WebElement Object.
        """
        self._execute(Command.SET_TIMEOUTS, {
            'ms': float(time_to_wait) * 1000,
            'type': 'page load'})

    @fluent
    def accept_alert(self):
        """Accepts the alert available.

        Support:
            iOS

        """
        self._execute(Command.ACCEPT_ALERT)

    @fluent
    def dismiss_alert(self):
        """Dismisses the alert available.

        Support:
            iOS
        """
        self._execute(Command.DISMISS_ALERT)

    @property
    def alert_text(self):
        """Gets the text of the Alert.

        Support:
            iOS
        """
        return self._execute(Command.GET_ALERT_TEXT)

    @fluent
    def alert_keys(self, keys):
        """Sends keystrokes to a JavaScript prompt() dialog.

        Support:
            iOS

        Args:
            keys(str): The keys send to.
        """
        self._execute(Command.SET_ALERT_VALUE, {
            'text': keys
        })

    def take_screenshot(self):
        """Gets the screenshot of the current window
           as a base64 encoded string.

        Support:
            Android iOS Web(WebView)

        Returns:
            Base64 encoded string of the screenshot.
        """
        return self._execute(Command.SCREENSHOT)

    @fluent
    def save_screenshot(self, filename, quietly = False):
        """Save the screenshot to local.

        Support:
            Android iOS Web(WebView)

        Args:
            filename(str): The path to save the image.
            quietly(bool): If True, omit the IOError when
                failed to save the image.

        Returns:
            WebElement Object.

        Raises:
            WebDriverException.
            IOError.
        """
        imgData = self.take_screenshot()
        try:
            with open(filename, "wb") as f:
                f.write(b64decode(imgData.encode('ascii')))
        except IOError as err:
            if not quietly:
                raise err

    def element(self, using, value):
        """Find an element in the current context.

        Support:
            Android iOS Web(WebView)

        Args:
            using(str): The element location strategy.
            value(str): The value of the location strategy.

        Returns:
            WebElement Object.

        Raises:
            WebDriverException.
        """
        return self._execute(Command.FIND_ELEMENT, {
            'using': using,
            'value': value
        })

    def element_if_exists(self, using, value):
        """Check if an element in the current context.

        Support:
            Android iOS Web(WebView)

        Args:
            using(str): The element location strategy.
            value(str): The value of the location strategy.

        Returns:
            Return True if the element does exists and return False otherwise.

        Raises:
            WebDriverException.
        """
        try:
            self._execute(Command.FIND_ELEMENT, {
                'using': using,
                'value': value
            })
            return True
        except:
            return False

    def element_or_none(self, using, value):
        """Check if an element in the current context.

        Support:
            Android iOS Web(WebView)

        Args:
            using(str): The element location strategy.
            value(str): The value of the location strategy.

        Returns:
            Return Element if the element does exists and return None otherwise.

        Raises:
            WebDriverException.
        """
        try:
            return self._execute(Command.FIND_ELEMENT, {
                'using': using,
                'value': value
            })
        except:
            return None

    def elements(self, using, value):
        """Find elements in the current context.

        Support:
            Android iOS Web(WebView)

        Args:
            using(str): The element location strategy.
            value(str): The value of the location strategy.

        Returns:
            Return a List<Element | None>, if no element matched, the list is empty.

        Raises:
            WebDriverException.
        """
        return self._execute(Command.FIND_ELEMENTS, {
            'using': using,
            'value': value
        })

    def wait_for(
        self, timeout=10000, interval=1000,
        asserter=lambda x: x):
        """Wait for driver till satisfy the given condition

        Support:
            Android iOS Web(WebView)

        Args:
            timeout(int): How long we should be retrying stuff.
            interval(int): How long between retries.
            asserter(callable): The asserter func to determine the result.

        Returns:
            Return the driver.

        Raises:
            WebDriverException.
        """
        if not callable(asserter):
            raise TypeError('Asserter must be callable.')
        @retry(
            retry_on_exception=lambda ex: isinstance(ex, WebDriverException),
            stop_max_delay=timeout,
            wait_fixed=interval
        )
        def _wait_for(driver):
            asserter(driver)
            return driver

        return _wait_for(self)

    def wait_for_element(
        self, using, value, timeout=10000,
        interval=1000, asserter=is_displayed):
        """Wait for element till satisfy the given condition

        Support:
            Android iOS Web(WebView)

        Args:
            using(str): The element location strategy.
            value(str): The value of the location strategy.
            timeout(int): How long we should be retrying stuff.
            interval(int): How long between retries.
            asserter(callable): The asserter func to determine the result.

        Returns:
            Return the Element.

        Raises:
            WebDriverException.
        """
        if not callable(asserter):
            raise TypeError('Asserter must be callable.')
        @retry(
            retry_on_exception=lambda ex: isinstance(ex, WebDriverException),
            stop_max_delay=timeout,
            wait_fixed=interval
        )
        def _wait_for_element(ctx, using, value):
            el = ctx.element(using, value)
            asserter(el)
            return el

        return _wait_for_element(self, using, value)

    def wait_for_elements(
        self, using, value, timeout=10000,
        interval=1000, asserter=is_displayed):
        """Wait for elements till satisfy the given condition

        Support:
            Android iOS Web(WebView)

        Args:
            using(str): The element location strategy.
            value(str): The value of the location strategy.
            timeout(int): How long we should be retrying stuff.
            interval(int): How long between retries.
            asserter(callable): The asserter func to determine the result.

        Returns:
            Return the list of Element if any of them satisfy the condition.

        Raises:
            WebDriverException.
        """
        if not callable(asserter):
            raise TypeError('Asserter must be callable.')
        @retry(
            retry_on_exception=lambda ex: isinstance(ex, WebDriverException),
            stop_max_delay=timeout,
            wait_fixed=interval
        )
        def _wait_for_elements(ctx, using, value):
            els = ctx.elements(using, value)
            if not len(els):
                raise WebDriverException('no such element')
            else:
                el = els[0]
                asserter(el)
                return els

        return _wait_for_elements(self, using, value)

    @fluent
    def touch(self, name, args=None):
        """Apply touch actions on devices. Such as, tap/doubleTap/press/pinch/rotate/drag.

        Support:
            Android iOS

        Args:
            name(str): Name of the action
            args(dict): Arguments of the action

        Returns:
            WebDriver Object.

        Raises:
            WebDriverException.
        """
        if isinstance(name, list) and not isinstance(name, str):
            actions = name
        elif isinstance(name, str):
            if not args:
                args = {}
            args['type'] = name
            actions = [args]
        else:
            raise TypeError('Invalid parameters.')
        self._execute(Command.PERFORM_ACTIONS, {
            'actions': actions
        })


add_element_extension_method(WebDriver)



#
# https = //w3c.github.io/webdriver/webdriver-spec.html
# WebDriver Exception
#

from collections import namedtuple
from enum import Enum

ErrorCode = namedtuple('ErrorCode', 'code error_code')


class WebDriverError(Enum):
    """
    Error codes defined in the WebDriver wire protocol.
    """
    NO_SUCH_ELEMENT = ErrorCode(7, 'no such element')
    NO_SUCH_FRAME = ErrorCode(8, 'no such frame')
    UNKNOWN_COMMAND = ErrorCode(9, 'unknown command')
    STALE_ELEMENT_REFERENCE = ErrorCode(10, 'stale element reference')
    ELEMENT_NOT_VISIBLE = ErrorCode(11, 'element not visible')
    INVALID_ELEMENT_STATE = ErrorCode(12, 'invalid element state')
    UNKNOWN_ERROR = ErrorCode(13, 'unknown error')
    ELEMENT_IS_NOT_SELECTABLE = ErrorCode(15, 'element not selectable')
    JAVASCRIPT_ERROR = ErrorCode(17, 'javascript error')
    XPATH_LOOKUP_ERROR = ErrorCode(19, 'invalid selector')
    TIMEOUT = ErrorCode(21, 'timeout')
    NO_SUCH_WINDOW = ErrorCode(23, 'no such window')
    INVALID_COOKIE_DOMAIN = ErrorCode(24, 'invalid cookie domain')
    UNABLE_TO_SET_COOKIE = ErrorCode(25, 'unable to set cookie')
    UNEXPECTED_ALERT_OPEN = ErrorCode(26, 'unexpected alert open')
    NO_ALERT_OPEN = ErrorCode(27, 'no such alert')
    SCRIPT_TIMEOUT = ErrorCode(28, 'script timeout')
    INVALID_ELEMENT_COORDINATES = ErrorCode(29, 'invalid element coordinates')
    IME_NOT_AVAILABLE = ErrorCode(30, 'ime not available')
    IME_ENGINE_ACTIVATION_FAILED = ErrorCode(
        31, 'ime engine activation failed')
    INVALID_SELECTOR = ErrorCode(32, 'invalid selector')
    MOVE_TARGET_OUT_OF_BOUNDS = ErrorCode(34, 'move target out of bounds')


def find_exception_by_code(code):
    """Find name of exception by WebDriver defined error code.

    Args:
        code(str): Error code defined in protocol.

    Returns:
        The error name defined in protocol.
    """
    errorName = None
    for error in WebDriverError:
        if error.value.code == code:
            errorName = error
            break
    return errorName


class WebDriverException(Exception):
    """WebDriver exception.

    Attributes:
        error(str): Error type defined in WebDriver Protocol.
        message(str): A description of the kind of error that occurred.
        screen(str):  If included, a screenshot of the current page as
            a base64 encoded string.
        stacktrace(str): A stack trace report of the error occurred.
    """

    def __init__(self, error=None, message=None, screen=None, stacktrace=None):
        """Initialize the WebDriverException"""
        self.error = error
        self.message = message
        self.screen = screen
        self.stacktrace = stacktrace

    def __str__(self):
        exception_msg = (
            "\nError: {0}\nMessage: {1}\n").format(self.error, self.message)
        if self.screen is not None:
            exception_msg += "Screenshot: available via screen\n"
        if self.stacktrace is not None:
            stacktrace = "\n".join(self.stacktrace)
            exception_msg += "Stacktrace:\n%s" % stacktrace
        return exception_msg



#
# https = //w3c.github.io/webdriver/webdriver-spec.html#list-of-endpoints
# Commands according to WebDriver Endpoints
#

from collections import namedtuple

Endpoint = namedtuple('Endpoint', ['method', 'uri'])


class Command(object):
    """Commands for WebDriver Defined Endpoints."""
    STATUS = Endpoint(
        'GET',
        '/status'
    )
    NEW_SESSION = Endpoint(
        'POST',
        '/session'
    )
    GET_ALL_SESSIONS = Endpoint(
        'GET',
        '/sessions'
    )
    QUIT = Endpoint(
        'DELETE',
        '/session/{session_id}'
    )
    GET_CURRENT_WINDOW_HANDLE = Endpoint(
        'GET',
        '/session/{session_id}/window_handle'
    )
    GET_WINDOW_HANDLES = Endpoint(
        'GET',
        '/session/{session_id}/window_handles'
    )
    GET = Endpoint(
        'POST',
        '/session/{session_id}/url'
    )
    GO_FORWARD = Endpoint(
        'POST',
        '/session/{session_id}/forward'
    )
    GO_BACK = Endpoint(
        'POST',
        '/session/{session_id}/back'
    )
    REFRESH = Endpoint(
        'POST',
        '/session/{session_id}/refresh'
    )
    EXECUTE_SCRIPT = Endpoint(
        'POST',
        '/session/{session_id}/execute'
    )
    GET_CURRENT_URL = Endpoint(
        'GET',
        '/session/{session_id}/url'
    )
    GET_TITLE = Endpoint(
        'GET',
        '/session/{session_id}/title'
    )
    GET_PAGE_SOURCE = Endpoint(
        'GET',
        '/session/{session_id}/source'
    )
    SCREENSHOT = Endpoint(
        'GET',
        '/session/{session_id}/screenshot'
    )
    ELEMENT_SCREENSHOT = Endpoint(
        'GET',
        '/session/{session_id}/element/{element_id}/screenshot'
    )
    FIND_ELEMENT = Endpoint(
        'POST',
        '/session/{session_id}/element'
    )
    FIND_ELEMENTS = Endpoint(
        'POST',
        '/session/{session_id}/elements'
    )
    GET_ACTIVE_ELEMENT = Endpoint(
        'POST',
        '/session/{session_id}/element/active'
    )
    FIND_CHILD_ELEMENT = Endpoint(
        'POST',
        '/session/{session_id}/element/{element_id}/element'
    )
    FIND_CHILD_ELEMENTS = Endpoint(
        'POST',
        '/session/{session_id}/element/{element_id}/elements'
    )
    CLICK_ELEMENT = Endpoint(
        'POST',
        '/session/{session_id}/element/{element_id}/click'
    )
    CLEAR_ELEMENT = Endpoint(
        'POST',
        '/session/{session_id}/element/{element_id}/clear'
    )
    SWIPE_ELEMENT = Endpoint(
        'POST',
        '/session/{session_id}/element/{element_id}/swipe'
    )
    SUBMIT_ELEMENT = Endpoint(
        'POST',
        '/session/{session_id}/element/{element_id}/submit'
    )
    GET_ELEMENT_TEXT = Endpoint(
        'GET',
        '/session/{session_id}/element/{element_id}/text'
    )
    SEND_KEYS_TO_ELEMENT = Endpoint(
        'POST',
        '/session/{session_id}/element/{element_id}/value'
    )
    SEND_KEYS_TO_ACTIVE_ELEMENT = Endpoint(
        'POST',
        '/session/{session_id}/keys'
    )
    UPLOAD_FILE = Endpoint(
        'POST',
        "/session/{session_id}/file"
    )
    GET_ELEMENT_VALUE = Endpoint(
        'GET',
        '/session/{session_id}/element/{element_id}/value'
    )
    GET_ELEMENT_TAG_NAME = Endpoint(
        'GET',
        '/session/{session_id}/element/{element_id}/name'
    )
    IS_ELEMENT_SELECTED = Endpoint(
        'GET',
        '/session/{session_id}/element/{element_id}/selected'
    )
    SET_ELEMENT_SELECTED = Endpoint(
        'POST',
        '/session/{session_id}/element/{element_id}/selected'
    )
    IS_ELEMENT_ENABLED = Endpoint(
        'GET',
        '/session/{session_id}/element/{element_id}/enabled'
    )
    IS_ELEMENT_DISPLAYED = Endpoint(
        'GET',
        '/session/{session_id}/element/{element_id}/displayed'
    )
    GET_ELEMENT_LOCATION = Endpoint(
        'GET',
        '/session/{session_id}/element/{element_id}/location'
    )
    GET_ELEMENT_LOCATION_ONCE_SCROLLED_INTO_VIEW = Endpoint(
        'GET',
        '/session/{session_id}/element/{element_id}/location_in_view'
    )
    GET_ELEMENT_SIZE = Endpoint(
        'GET',
        '/session/{session_id}/element/{element_id}/size'
    )
    GET_ELEMENT_RECT = Endpoint(
        'GET',
        '/session/{session_id}/element/{element_id}/rect'
    )
    GET_ELEMENT_PROPERTY = Endpoint(
        'GET',
        '/session/{session_id}/element/{element_id}/property/{name}'
    )
    GET_ELEMENT_ATTRIBUTE = Endpoint(
        'GET',
        '/session/{session_id}/element/{element_id}/attribute/{name}'
    )
    ELEMENT_EQUALS = Endpoint(
        'GET',
        '/session/{session_id}/element/{element_id}/equals/{other}'
    )
    GET_ALL_COOKIES = Endpoint(
        'GET',
        '/session/{session_id}/cookie'
    )
    ADD_COOKIE = Endpoint(
        'POST',
        '/session/{session_id}/cookie'
    )
    DELETE_ALL_COOKIES = Endpoint(
        'DELETE',
        '/session/{session_id}/cookie'
    )
    DELETE_COOKIE = Endpoint(
        'DELETE',
        '/session/{session_id}/cookie/{name}'
    )
    SWITCH_TO_FRAME = Endpoint(
        'POST',
        '/session/{session_id}/frame'
    )
    SWITCH_TO_PARENT_FRAME = Endpoint(
        'POST',
        '/session/{session_id}/frame/parent'
    )
    SWITCH_TO_WINDOW = Endpoint(
        'POST',
        '/session/{session_id}/window'
    )
    CLOSE = Endpoint(
        'DELETE',
        '/session/{session_id}/window'
    )
    GET_ELEMENT_VALUE_OF_CSS_PROPERTY = Endpoint(
        'GET',
        '/session/{session_id}/element/{element_id}/css/{property_name}'
    )
    IMPLICIT_WAIT = Endpoint(
        'POST',
        '/session/{session_id}/timeouts/implicit_wait'
    )
    EXECUTE_ASYNC_SCRIPT = Endpoint(
        'POST',
        '/session/{session_id}/execute_async'
    )
    SET_SCRIPT_TIMEOUT = Endpoint(
        'POST',
        '/session/{session_id}/timeouts/async_script'
    )
    SET_TIMEOUTS = Endpoint(
        'POST',
        '/session/{session_id}/timeouts'
    )
    DISMISS_ALERT = Endpoint(
        'POST',
        '/session/{session_id}/dismiss_alert'
    )
    ACCEPT_ALERT = Endpoint(
        'POST',
        '/session/{session_id}/accept_alert'
    )
    SET_ALERT_VALUE = Endpoint(
        'POST',
        '/session/{session_id}/alert_text'
    )
    GET_ALERT_TEXT = Endpoint(
        'GET',
        '/session/{session_id}/alert_text'
    )
    SET_ALERT_CREDENTIALS = Endpoint(
        'POST',
        '/session/{session_id}/alert/credentials'
    )
    CLICK = Endpoint(
        'POST',
        '/session/{session_id}/click'
    )
    DOUBLE_CLICK = Endpoint(
        'POST',
        '/session/{session_id}/doubleclick'
    )
    MOUSE_DOWN = Endpoint(
        'POST',
        '/session/{session_id}/buttondown'
    )
    MOUSE_UP = Endpoint(
        'POST',
        '/session/{session_id}/buttonup'
    )
    MOVE_TO = Endpoint(
        'POST',
        '/session/{session_id}/moveto'
    )
    GET_WINDOW_SIZE = Endpoint(
        'GET',
        '/session/{session_id}/window/{window_handle}/size'
    )
    W3C_GET_WINDOW_SIZE = Endpoint(
        'GET',
        '/session/{session_id}/window/size'
    )
    SET_WINDOW_SIZE = Endpoint(
        'POST',
        '/session/{session_id}/window/{window_handle}/size'
    )
    W3C_SET_WINDOW_SIZE = Endpoint(
        'POST',
        '/session/{session_id}/window/size'
    )
    GET_WINDOW_POSITION = Endpoint(
        'GET',
        '/session/{session_id}/window/{window_handle}/position'
    )
    SET_WINDOW_POSITION = Endpoint(
        'POST',
        '/session/{session_id}/window/{window_handle}/position'
    )
    MAXIMIZE_WINDOW = Endpoint(
        'POST',
        '/session/{session_id}/window/{window_handle}/maximize'
    )
    W3C_MAXIMIZE_WINDOW = Endpoint(
        'POST',
        '/session/{session_id}/window/maximize'
    )
    SET_SCREEN_ORIENTATION = Endpoint(
        'POST',
        '/session/{session_id}/orientation'
    )
    GET_SCREEN_ORIENTATION = Endpoint(
        'GET',
        '/session/{session_id}/orientation'
    )
    SINGLE_TAP = Endpoint(
        'POST',
        '/session/{session_id}/touch/click'
    )
    TOUCH_DOWN = Endpoint(
        'POST',
        '/session/{session_id}/touch/down'
    )
    TOUCH_UP = Endpoint(
        'POST',
        '/session/{session_id}/touch/up'
    )
    TOUCH_MOVE = Endpoint(
        'POST',
        '/session/{session_id}/touch/move'
    )
    TOUCH_SCROLL = Endpoint(
        'POST',
        '/session/{session_id}/touch/scroll'
    )
    DOUBLE_TAP = Endpoint(
        'POST',
        '/session/{session_id}/touch/doubleclick'
    )
    LONG_PRESS = Endpoint(
        'POST',
        '/session/{session_id}/touch/longclick'
    )
    FLICK = Endpoint(
        'POST',
        '/session/{session_id}/touch/flick'
    )
    EXECUTE_SQL = Endpoint(
        'POST',
        '/session/{session_id}/execute_sql'
    )
    GET_LOCATION = Endpoint(
        'GET',
        '/session/{session_id}/location'
    )
    SET_LOCATION = Endpoint(
        'POST',
        '/session/{session_id}/location'
    )
    GET_APP_CACHE = Endpoint(
        'GET',
        '/session/{session_id}/application_cache'
    )
    GET_APP_CACHE_STATUS = Endpoint(
        'GET',
        '/session/{session_id}/application_cache/status'
    )
    CLEAR_APP_CACHE = Endpoint(
        'DELETE',
        '/session/{session_id}/application_cache/clear'
    )
    GET_NETWORK_CONNECTION = Endpoint(
        'GET',
        '/session/{session_id}/network_connection'
    )
    SET_NETWORK_CONNECTION = Endpoint(
        'POST',
        '/session/{session_id}/network_connection'
    )
    GET_LOCAL_STORAGE_ITEM = Endpoint(
        'GET',
        '/session/{session_id}/local_storage/key/{key}'
    )
    REMOVE_LOCAL_STORAGE_ITEM = Endpoint(
        'DELETE',
        '/session/{session_id}/local_storage/key/{key}'
    )
    GET_LOCAL_STORAGE_KEYS = Endpoint(
        'GET',
        '/session/{session_id}/local_storage'
    )
    SET_LOCAL_STORAGE_ITEM = Endpoint(
        'POST',
        '/session/{session_id}/local_storage'
    )
    CLEAR_LOCAL_STORAGE = Endpoint(
        'DELETE',
        '/session/{session_id}/local_storage'
    )
    GET_LOCAL_STORAGE_SIZE = Endpoint(
        'GET',
        '/session/{session_id}/local_storage/size'
    )
    GET_SESSION_STORAGE_ITEM = Endpoint(
        'GET',
        '/session/{session_id}/session_storage/key/{key}'
    )
    REMOVE_SESSION_STORAGE_ITEM = Endpoint(
        'DELETE',
        '/session/{session_id}/session_storage/key/{key}'
    )
    GET_SESSION_STORAGE_KEYS = Endpoint(
        'GET',
        '/session/{session_id}/session_storage'
    )
    SET_SESSION_STORAGE_ITEM = Endpoint(
        'POST',
        '/session/{session_id}/session_storage'
    )
    CLEAR_SESSION_STORAGE = Endpoint(
        'DELETE',
        '/session/{session_id}/session_storage'
    )
    GET_SESSION_STORAGE_SIZE = Endpoint(
        'GET',
        '/session/{session_id}/session_storage/size'
    )
    GET_LOG = Endpoint(
        'POST',
        '/session/{session_id}/log'
    )
    GET_AVAILABLE_LOG_TYPES = Endpoint(
        'GET',
        '/session/{session_id}/log/types'
    )
    CURRENT_CONTEXT_HANDLE = Endpoint(
        'GET',
        '/session/{session_id}/context'
    )
    CONTEXT_HANDLES = Endpoint(
        'GET',
        '/session/{session_id}/contexts'
    )
    SWITCH_TO_CONTEXT = Endpoint(
        'POST',
        '/session/{session_id}/context'
    )
    PERFORM_ACTIONS = Endpoint(
        'POST',
        '/session/{session_id}/actions'
    )



#
# Remote Invoker to execute command & handle HTTP communication
#

import logging
try:
    from urllib.parse import urlparse, urlunparse
except ImportError:
    from urlparse import urlparse, urlunparse

from requests import Request, Session

from .util import MemorizeFormatter

LOGGER = logging.getLogger(__name__)


class RemoteInvoker(object):
    """Remote Invoker to execute WebDriver command."""

    def __init__(self, url='http://127.0.0.1:3456/wd/hub'):
        """Init the RemoteInvoker by remote url

        Args:
            url(str|dict): The url of remote server.
        Defaults:
            if url is str:
                url = http://127.0.0.1:3456/wd/hub
            if url is dict:
                url = {
                    'protocol': 'http',
                    'hostname': '127.0.0.1',
                    'port': 3456,
                    'path': '/wd/hub'
                }
        Examples:
            r = RemoteInvoker('http://127.0.0.1:3456/wd/hub')
            r = RemoteInvoker({
                'protocol': 'https',
                'hostname': '127.0.0.1',
                'port': 5678,
                'username': 'macaca',
                'password': '123456',
                'path': '/how/r/u'
            }) => "https://macaca:123456@127.0.0.1:5678/how/r/u"
        """
        self._timeout = None
        if isinstance(url, str):
            parsed_url = urlparse(url)
            scheme = parsed_url.scheme
            netloc = parsed_url.netloc
            if not scheme or not netloc:
                raise ValueError(
                    'Invalid URL \'{0}\': No schema or '
                    'hostname supplied'.format(url))
            elif scheme not in ('http', 'https'):
                raise ValueError(
                    'Invalid URL \'{0}\': Unknown schema \'{1}\', '
                    'only \'http\' and \'https\' '
                    'are supported'.format(url, scheme))
            else:
                self._url = url
        elif isinstance(url, dict):
            scheme = url.get('scheme', None) \
                or url.get('protocol', None) \
                or 'http'
            if scheme not in ('http', 'https'):
                raise ValueError(
                    'Invalid URL Dict: Unknown schema \'{0}\', '
                    'only \'http\' and \'https\' are supported'.format(scheme))
            hostname = url.get('hostname', '127.0.0.1')
            port = url.get('port', 3456)
            path = url.get('path', '/wd/hub')
            username = url.get('username', None)
            password = url.get('password', None)
            if username and password:
                netloc = '{0}:{1}@{2}:{3}'.format(
                    username, password, hostname, port)
            else:
                netloc = '{0}:{1}'.format(hostname, port)
            self._url = urlunparse((scheme, netloc, path, '', '', ''))
        else:
            raise TypeError(
                'URL \'{0}\' should be a string or a dict.'.format(url))

        self._formatter = MemorizeFormatter()

    def execute(self, command, data={}):
        """Format the endpoint url by data and then request the remote server.

        Args:
            command(Command): WebDriver command to be executed.
            data(dict): Data fulfill the uri template and json body.

        Returns:
            A dict represent the json body from server response.

        Raises:
            KeyError: Data cannot fulfill the variable which command needed.
            ConnectionError: Meet network problem (e.g. DNS failure,
                refused connection, etc).
            Timeout: A request times out.
            HTTPError: HTTP request returned an unsuccessful status code.
        """
        method, uri = command
        try:
            path = self._formatter.format_map(uri, data)
            body = self._formatter.get_unused_kwargs()
            url = "{0}{1}".format(self._url, path)
            return self._request(method, url, body)
        except KeyError as err:
            LOGGER.debug(
                'Endpoint {0} is missing argument {1}'.format(uri, err))
            raise

    def _request(self, method, url, body):
        """Internal method to send request to the remote server.

        Args:
            method(str): HTTP Method(GET/POST/PUT/DELET/HEAD).
            url(str): The request url.
            body(dict): The JSON object to be sent.

        Returns:
            A dict represent the json body from server response.

        Raises:
            ConnectionError: Meet network problem (e.g. DNS failure,
                refused connection, etc).
            Timeout: A request times out.
            HTTPError: HTTP request returned an unsuccessful status code.
        """
        if method != 'POST' and method != 'PUT':
            body = None

        s = Session()

        LOGGER.debug(
            'Method: {0}, Url: {1}, Body: {2}.'.format(method, url, body))

        req = Request(method, url, json=body)
        prepped = s.prepare_request(req)

        res = s.send(prepped, timeout=self._timeout or None)
        res.raise_for_status()
        # TODO try catch
        return res.json()



#
# Keys for WebDriver
#

from enum import Enum


class Keys(Enum):
    """The Unicode Private Use Area code points, 0xE000-0xF8FF,
       are used to represent pressable, non-text keys.
    """

    NULL = '\ue000'
    CANCEL = '\ue001'
    HELP = '\ue002'
    BACKSPACE = '\ue003'
    TAB = '\ue004'
    CLEAR = '\ue005'
    RETURN = '\ue006'
    ENTER = '\ue007'
    SHIFT = '\ue008'
    CONTROL = '\ue009'
    ALT = '\ue00a'
    PAUSE = '\ue00b'
    ESCAPE = '\ue00c'
    SPACE = '\ue00d'
    PAGE_UP = '\ue00e'
    PAGE_DOWN = '\ue00f'
    END = '\ue010'
    HOME = '\ue011'
    ARROW_LEFT = '\ue012'
    ARROW_UP = '\ue013'
    ARROW_RIGHT = '\ue014'
    ARROW_DOWN = '\ue015'
    INSERT = '\ue016'
    DELETE = '\ue017'
    SEMICOLON = '\ue018'
    EQUALS = '\ue019'
    NUMPAD0 = '\ue01a'
    NUMPAD1 = '\ue01b'
    NUMPAD2 = '\ue01c'
    NUMPAD3 = '\ue01d'
    NUMPAD4 = '\ue01e'
    NUMPAD5 = '\ue01f'
    NUMPAD6 = '\ue020'
    NUMPAD7 = '\ue021'
    NUMPAD8 = '\ue022'
    NUMPAD9 = '\ue023'
    MULTIPLY = '\ue024'
    ADD = '\ue025'
    SEPARATOR = '\ue026'
    SUBTRACT = '\ue027'
    DECIMAL = '\ue028'
    DIVIDE = '\ue029'
    F1 = '\ue031'
    F2 = '\ue032'
    F3 = '\ue033'
    F4 = '\ue034'
    F5 = '\ue035'
    F6 = '\ue036'
    F7 = '\ue037'
    F8 = '\ue038'
    F9 = '\ue039'
    F10 = '\ue03a'
    F11 = '\ue03b'
    F12 = '\ue03c'
    META = '\ue03d'
    COMMAND = '\ue03d'
    ZENKAKU_HANKAKU	= '\ue040'



#
# https = //w3c.github.io/webdriver/webdriver-spec.html
# WebDriver Element Implemenation
#

from base64 import b64decode

from retrying import retry

from .asserters import is_displayed
from .command import Command
from .locator import Locator
from .util import add_element_extension_method, value_to_key_strokes, fluent
from .webdriverexception import WebDriverException


class WebElement(object):
    """The WebElement Object to implement most part of WebDriver protocol.

    Attributes:
        element_id(str): A UDID used to uniquely identify an element.
    """

    def __init__(self, element_id, driver):
        """Initialize the WebElement

        Args:
            element_id(str): The UDID returned by remote servers.
            driver(WebDriver): The WebDriver Object.
        """
        self.element_id = str(element_id)
        self._driver = driver

    def __repr__(self):
        return '<{0.__name__} (session="{1}", element="{2}")>'.format(
            type(self), self._driver.session_id, self.element_id)

    def __eq__(self, el):
        return hasattr(el, 'element_id') and self.element_id == el.element_id

    def __ne__(self, el):
        return not self.__eq__(el)

    def __hash__(self):
        return hash(self.element_id)

    def _execute(self, command, data=None, unpack=True):
        """Private method to execute command with data.

        Args:
            command(Command): The defined command.
            data(dict): The uri variable and body.

        Returns:
            The unwrapped value field in the json response.
        """
        if not data:
            data = {}
        data.setdefault('element_id', self.element_id)
        return self._driver._execute(command, data, unpack)

    @property
    def driver(self):
        """Internal reference to the WebDriver instance."""
        return self._driver

    def element(self, using, value):
        """find an element in the current element.

        Support:
            Android iOS Web(WebView)

        Args:
            using(str): The element location strategy.
            value(str): The value of the location strategy.

        Returns:
            WebElement Object.

        Raises:
            WebDriverException.
        """
        return self._execute(Command.FIND_CHILD_ELEMENT, {
            'using': using,
            'value': value
        })

    def element_if_exists(self, using, value):
        """Check if an element in the current element.

        Support:
            Android iOS Web(WebView)

        Args:
            using(str): The element location strategy.
            value(str): The value of the location strategy.

        Returns:
            Return True if the element does exists and return False otherwise.

        Raises:
            WebDriverException.
        """
        try:
            self._execute(Command.FIND_CHILD_ELEMENT, {
                'using': using,
                'value': value
            })
            return True
        except:
            return False

    def element_or_none(self, using, value):
        """Check if an element in the current element.

        Support:
            Android iOS Web(WebView)

        Args:
            using(str): The element location strategy.
            value(str): The value of the location strategy.

        Returns:
            Return Element if the element does exists and return None otherwise.

        Raises:
            WebDriverException.
        """
        try:
            return self._execute(Command.FIND_CHILD_ELEMENT, {
                'using': using,
                'value': value
            })
        except:
            return None

    def elements(self, using, value):
        """find elements in the current element.

        Support:
            Android iOS Web(WebView)

        Args:
            using(str): The element location strategy.
            value(str): The value of the location strategy.

        Returns:
            Return a List<Element | None>, if no element matched, the list is empty.

        Raises:
            WebDriverException.
        """
        return self._execute(Command.FIND_CHILD_ELEMENTS, {
            'using': using,
            'value': value
        })

    def wait_for(
        self, timeout=10000, interval=1000,
        asserter=lambda x: x):
        """Wait for element till given condition

        Support:
            Android iOS Web(WebView)

        Args:
            timeout(int): How long we should be retrying stuff.
            interval(int): How long between retries.
            asserter(callable): The asserter func to determine the result.

        Returns:
            Return the Element.

        Raises:
            WebDriverException.
        """
        if not callable(asserter):
            raise TypeError('Asserter must be callable.')
        @retry(
            retry_on_exception=lambda ex: isinstance(ex, WebDriverException),
            stop_max_delay=timeout,
            wait_fixed=interval
        )
        def _wait_for(el):
            asserter(el)
            return el

        return _wait_for(self)

    def wait_for_element(
        self, using, value, timeout=10000,
        interval=1000, asserter=is_displayed):
        """Wait for element till the given condition

        Support:
            Android iOS Web(WebView)

        Args:
            using(str): The element location strategy.
            value(str): The value of the location strategy.
            timeout(int): How long we should be retrying stuff.
            interval(int): How long between retries.
            asserter(callable): The asserter func to determine the result.

        Returns:
            Return the Element.

        Raises:
            WebDriverException.
        """
        if not callable(asserter):
            raise TypeError('Asserter must be callable.')
        @retry(
            retry_on_exception=lambda ex: isinstance(ex, WebDriverException),
            stop_max_delay=timeout,
            wait_fixed=interval
        )
        def _wait_for_element(ctx, using, value):
            el = ctx.element(using, value)
            asserter(el)
            return el

        return _wait_for_element(self, using, value)

    def wait_for_elements(
        self, using, value, timeout=10000,
        interval=1000, asserter=is_displayed):
        """Wait for elements till the given condition

        Support:
            Android iOS Web(WebView)

        Args:
            using(str): The element location strategy.
            value(str): The value of the location strategy.
            timeout(int): How long we should be retrying stuff.
            interval(int): How long between retries.
            asserter(callable): The asserter func to determine the result.

        Returns:
            Return the list of Element if any of them satisfy the condition.

        Raises:
            WebDriverException.
        """
        if not callable(asserter):
            raise TypeError('Asserter must be callable.')
        @retry(
            retry_on_exception=lambda ex: isinstance(ex, WebDriverException),
            stop_max_delay=timeout,
            wait_fixed=interval
        )
        def _wait_for_elements(ctx, using, value):
            els = ctx.elements(using, value)
            if not len(els):
                raise WebDriverException('no such element')
            else:
                el = els[0]
                asserter(el)
                return els

        return _wait_for_elements(self, using, value)

    def is_displayed(self):
        """Whether the element is visible.

        Support:
            Android Web(WebView)
        """
        return self._execute(Command.IS_ELEMENT_DISPLAYED)

    def is_selected(self):
        """Returns whether the element is selected.

        Support:
            Web(WebView)
        """
        return self._execute(Command.IS_ELEMENT_SELECTED)

    def is_enabled(self):
        """Returns whether the element is enabled.

        Support:
            Web(WebView)
        """
        return self._execute(Command.IS_ELEMENT_ENABLED)

    def get_property(self, name):
        """Return the result of getting a property, Support: Android iOS Web(WebView).

        Support:
            Support: Android iOS Web(WebView)

        Args:
            name(str): Name of the property to retrieve, Android iOS can only get `origin` `size`.

        Returns:
            The property of the element.
        """
        return self._execute(Command.GET_ELEMENT_PROPERTY, {'name': name})

    def get_computed_css(self, property_name):
        """The computed value of the given CSS property
           of the given web element.

        Support:
            Support: Web(WebView)

        Args:
            property_name(str): CSS property.

        Returns:
            The computed value of parameter property name
            from element's style declarations if the current
            browsing context's document type is not "xml",
            else let it be ""
        """
        return self._execute(Command.GET_ELEMENT_VALUE_OF_CSS_PROPERTY, {
            'property_name': property_name})

    @property
    def text(self):
        """Return the text of the element.
           This is equivalent to calling element.innerText.

        Support:
            Android iOS Web(WebView)
        """
        return self._execute(Command.GET_ELEMENT_TEXT)

    @property
    def tag_name(self):
        """Return the tagName of the element.

        Support:
            Web(WebView)
        """
        return self._execute(Command.GET_ELEMENT_TAG_NAME)

    @property
    def location(self):
        """The location of the element in the renderable canvas in pixels.

        Support:
            Web(WebView)

        Returns:
            A dict contains:
            x(float): X axis position of the top-left corner.
            y(float): Y axis position of the top-left corner.
        """
        return self._execute(Command.GET_ELEMENT_LOCATION)

    @property
    def rect(self):
        """The dimensions and coordinates of the given web element in pixels.

        Support:
            Web(WebView)

        Returns:
            A dict contains:
            x(float): X axis position of the top-left corner.
            y(float): Y axis position of the top-left corner.
            height(float): Height of the web element's bounding rectangle.
            width(float): Width of the web element's bounding rectangle.
        """
        return self._execute(Command.GET_ELEMENT_RECT)

    @property
    def size(self):
        """The size of the given web element in pixels.

        Support:
            Web(WebView)

        Returns:
            A dict contains:
            height(float): Height of the web element's bounding rectangle.
            width(float): Width of the web element's bounding rectangle.
        """
        return self._execute(Command.GET_ELEMENT_SIZE)

    @fluent
    def click(self):
        """The Element Click command scrolls into view
           the element and then attempts to click the
           centre of the visible area of the first element
           of the DOMRect sequence. In case the element is
           not displayed, an element not visible error is returned.

        Support:
            Android iOS Web(WebView)
        """
        self._execute(Command.CLICK_ELEMENT)

    @fluent
    def clear(self):
        """The Element Clear command scrolls into view
           a submittable element excluding buttons or
           editable element, and then attempts to clear
           its value, checkedness, or text content.

        Support:
            Android iOS Web(WebView)
        """
        self._execute(Command.CLEAR_ELEMENT)

    @fluent
    def send_keys(self, value):
        """Send a sequence of key strokes to an element.

        Support:
            Android iOS Web(WebView)

        Args:
            value(str|int|list): value can be a string,
              int or a list contains defined Keys.
        """
        self._execute(Command.SEND_KEYS_TO_ELEMENT, {
            'value': value_to_key_strokes(value)})

    @fluent
    def move_to(self, x=0, y=0):
        """Move the mouse by an offset of the specificed element.

        Support:
            Android

        Args:
            x(float): X offset to move to, relative to the
                      top-left corner of the element.
            y(float): Y offset to move to, relative to the
                      top-left corner of the element.

        Returns:
            WebElement object.
        """
        self._driver.move_to(self, x, y)

    @fluent
    def flick(self, x, y, speed):
        """Flick on the touch screen using finger motion events.
           This flickcommand starts at a particulat screen location.

        Support:
            iOS

        Args:
            x(float}: The x offset in pixels to flick by.
            y(float): The y offset in pixels to flick by.
            speed(float) The speed in pixels per seconds.

        Returns:
            WebElement object.
        """
        self._driver.flick(self, x, y, speed)

    @fluent
    def tap(self):
        """Single tap on the touch enabled device.

        Support:
            Android iOS

        Args:
            element(WebElement): WebElement Object to single tap on.

        Returns:
            WebElement object.
        """
        self._driver.tap(self)

    @fluent
    def swipe(self, startX, startY, endX, endY, duration = 1000):
        """Swipe on the touch screen using finger motion events.

        Support:
            Android iOS

        Args:
            startX(Float): The x-coordinate of start point.
            startY(Float): The y-coordinate of start point.
            endX(Float): The x-coordinate of end point.
            endY(Float): The y-coordinate of end point.
            duration(Float): The duration in ms.

        Returns:
            WebElement object.
        """
        self._execute(Command.SWIPE_ELEMENT, {
            'startX': startX,
            'startY': startY,
            'endX': endX,
            'endY': endY,
            'duration': duration
        })

    def take_screenshot(self):
        """Gets the screenshot of the current element
           as a base64 encoded string.

        Support:
            Android iOS Web(WebView)

        Returns:
            Base64 encoded string of the screenshot.
        """
        return self._execute(Command.ELEMENT_SCREENSHOT)

    @fluent
    def save_screenshot(self, filename, quietly = False):
        """Save the screenshot to local.

        Support:
            Android iOS Web(WebView)

        Args:
            filename(str): The path to save the image.
            quietly(bool): If True, omit the IOError when
                failed to save the image.

        Returns:
            WebElement Object.

        Raises:
            WebDriverException.
            IOError.
        """
        imgData = self.take_screenshot()
        try:
            with open(filename, "wb") as f:
                f.write(b64decode(imgData.encode('ascii')))
        except IOError as err:
            if not quietly:
                raise err

    @fluent
    def touch(self, name, args=None):
        """Apply touch actions on devices. Such as, tap/doubleTap/press/pinch/rotate/drag.

        Support:
            Android iOS

        Args:
            name(str): Name of the action
            args(dict): Arguments of the action

        Returns:
            WebDriver Object.

        Raises:
            WebDriverException.
        """
        if isinstance(name, list) and not isinstance(name, str):
            for obj in name:
                obj['element'] = self.element_id
            actions = name
        elif isinstance(name, str):
            if not args:
                args = {}
            args['type'] = name
            args['element'] = self.element_id
            actions = [args]
        else:
            raise TypeError('Invalid parameters.')
        self._driver._execute(Command.PERFORM_ACTIONS, {
            'actions': actions
        })


add_element_extension_method(WebElement)



#
# https://w3c.github.io/webdriver/webdriver-spec.html#locator-strategies
# Locator according to WebDriver Protocol
#

from enum import Enum


class Locator(Enum):
    """Locator Enum defined by WebDriver Protocol."""
    ID = "id"
    XPATH = "xpath"
    LINK_TEXT = "link text"
    PARTIAL_LINK_TEXT = "partial link text"
    NAME = "name"
    TAG_NAME = "tag name"
    CLASS_NAME = "class name"
    CSS_SELECTOR = "css selector"
